"""Constants used for the servers, clients and brokers"""

PSERVER = 'QWPS01'.encode() #Server following majordomopatternv 0.1
PCLIENT = 'QWPC01'.encode() #Client following majordomopatternv 0.1
CREADY = '1'.encode()
CREQUEST = '2'.encode()
CREPLY = '3'.encode()
CDISCONNECT = '4'.encode()

CFAIL = '00'.encode()
CSUCCESS = '11'.encode()

CHEARTBEAT = '42'.encode()

HEARTBEATMAX = 5


from .constants import *
import zmq
import pickle
import time
def QMethod(func):
    '''Decorator for exposing methods that can be called by clients'''
    func.is_client_accessible = True
    return func


class QWeatherServer:
    def initialize_sockets(self):
        if self.verbose:
            print('#########')
            print('Connecting "',self.servername,'" to QWeatherStation on IP "',self.QWeatherStationIP,'"\n')
        self.servername = self.servername.encode()
        self.context = zmq.Context()
        self.QWeatherStation = self.context.socket(zmq.DEALER)
        self.QWeatherStation.connect(self.QWeatherStationIP)
        if self.verbose:
            print('Connection established\n')

#        self.broadcastsocket = self.context.socket(zmq.PUB)
 #       self.broadcastsocket.connect(broadcastsocket)

        self.poller = zmq.Poller()
        self.poller.register(self.QWeatherStation,zmq.POLLIN)

        self.methoddict = {func:getattr(self,func) for func in dir(self) if getattr(getattr(self,func),'is_client_accessible',False)}
        self.register_at_station()


    def register_at_station(self):
        methodlist = [(func,getattr(self,func).__doc__) for func in dir(self) if getattr(getattr(self,func),'is_client_accessible',False)]
        msg = [b'',b'S',CREADY,PSERVER,self.servername,pickle.dumps(methodlist)]
        if self.debug:
            print('DEBUG: To QWeatherStation: ',msg)
        self.QWeatherStation.send_multipart(msg)

    def run(self):
        #Broadcast method information
        tic = time.time()
        while True:
            try:
                items = self.poller.poll(1000)
                if items:
                    msg = self.QWeatherStation.recv_multipart()
                    self.handle_messages(msg)
                toc = time.time()
                if toc-tic > 1:
                    answ = [b'',b'H',self.servername]
                    self.QWeatherStation.send_multipart(answ)    
                    tic = toc
            except KeyboardInterrupt:
                break


    def handle_messages(self,msg):
        if self.debug:
            print('DEBUG: From QWeatherStation: ',msg)
        empty = msg.pop(0)
        assert empty == b''
        command = msg.pop(0)

        if command == CREQUEST:
            client = msg.pop(0)
            fnc = msg.pop(0).decode()
            args,kwargs = pickle.loads(msg.pop(0))
            if self.debug:
                print('DEBUG: Calling function: ',fnc,' with arguments: ',args,kwargs)
            answ = self.methoddict[fnc](*args,**kwargs)
            answ = [empty,b'S',CREPLY] + [self.servername,client,pickle.dumps(answ)]
            if self.debug:
                print('DEBUG: To QWeatherStation: ', answ)
            self.QWeatherStation.send_multipart(answ)        

        elif command == CREADY+CSUCCESS:
            if self.verbose:
                print('\nMethods registered at QWeatherStation. (',[i[0] for i in methodlist],')')

        elif command == CREADY+CFAIL:
            raise Exception(msg.pop(0).decode())
        #if command == 

        

    def initialize_broadcasting(self):
        if self.debug:
            print('DEBUG: Initializing broadcasting')



from .constants import *
import zmq
import pickle
from zmq.asyncio import Context, Poller
import asyncio
import time
class QWeatherClient:

    class serverclass:
        def __init__(self,name,addr,methods,client):
            self.name = name
            self.addr = addr
            self.client = client
            for amethod in methods:
                setattr(self,amethod[0],self.bindingfunc(amethod[0],amethod[1]))


        def bindingfunc(self,methodname,methoddoc):
            def func(*args,**kwargs):
                return self.client.send_request([self.name.encode(),methodname.encode(),pickle.dumps([args,kwargs])])
            func.__name__ = methodname
            func.__doc__ = methoddoc
            func.is_remote_server_method = True
            return func


        def __repr__(self):
            msg = ""
            lst = [getattr(self,method) for method in dir(self) if getattr(getattr(self,method),'is_remote_server_method',False)]
            if len(lst) == 0:
                return 'No servers connected'
            else:
                for amethod in lst:
                    msg += amethod.__name__ +"\n"
            return msg.strip()
    

    context = None
    socket = None
    poller = None
    futureobjectdict = {}

    def __init__(self,QWeatherStationIP,name = None,loop = None):
        self.QWeatherStationIP = QWeatherStationIP
        if loop is None:
            self.loop = asyncio.get_event_loop()
        else:
            self.loop = loop

        if name is None:
            import socket
            name = socket.gethostname()
        self.name = name.encode()
        self.reconnect()
        self.loop.run_until_complete(self.get_server_info())
        self.running = False



    def reconnect(self):
        '''connects or reconnects to the broker'''
        if self.poller:
            self.poller.unregister(self.socket)
        if self.socket: 
            self.socket.close()
        if self.context:
            self.context.term()
        self.context = Context()
        self.socket = self.context.socket(zmq.DEALER)
        self.socket.connect(self.QWeatherStationIP)
        self.serverlist = []
        self.poller = Poller()
        self.poller.register(self.socket,zmq.POLLIN)
        #lol = asyncio.ensure_future()
        
    
    async def get_server_info(self):
        msg = [b'',b'C',CREADY,PCLIENT,self.name]
        self.send_message(msg)
        msg =  await self.socket.recv_multipart()
        empty = msg.pop(0)
        assert empty == b''
        command = msg.pop(0)
        if command == CREADY + CFAIL:
            raise Exception(msg.pop(0).decode())
        else:
            for name,items in pickle.loads(msg.pop(0)).items():
                addr = items[0]
                methods = items[1]
                server = self.serverclass(name,addr,methods,self)
                server.is_remote_server = True
                setattr(self,name,server)
            

#                _method = 
 #               _method = methodclass(self.send_request,server.name,amethod[0],amethod[1])
  #              _method.is_remote_server_method = True
   #             setattr(server,amethod[0],_method)
        #print('loaded servers')
        return None

    def send_request(self,body):
        if self.running:
            result =  asyncio.get_event_loop().create_task(self.async_send_request(body))
        else:
            result = self.sync_send_request(body)
        return result

    def sync_send_request(self,body):
        msg = [b'',b'C',CREQUEST] + body
        server = body[0]
        self.send_message(msg)
        msg = self.loop.run_until_complete(self.socket.recv_multipart())
        empty = msg.pop(0)
        assert empty == b''
        command = msg.pop(0)

        server = msg.pop(0)
        answ = pickle.loads(msg[0])
        return answ
    
    async def async_send_request(self,body):
        msg = [b'',b'C',CREQUEST] + body
        server = body[0]
        #print(body)
        if server in self.futureobjectdict.keys():
            raise Exception('Already waiting for response from that server')
        self.send_message(msg)
        answ = await self.recieve_message(server)
        self.futureobjectdict.pop(server)
        return answ
       # return msg

    def send_message(self,msg):
        #self.loop.create_task(self.socket.send_multipart(msg))
        self.socket.send_multipart(msg)


    def recieve_message(self,servername):
        tmp = self.loop.create_future()
        self.futureobjectdict[servername] = tmp
        #print('went here')
        return tmp

        #ans =self.loop.create_task(self.socket.recv_multipart())

#        ans = await self.socket.recv_multipart()
  #      return ans

    async def run(self):
        self.running = True
        tic = time.time()
        while True:
            try:
                #print('polling')
                items = await self.poller.poll(1000)
                if items:
                    msg = await self.socket.recv_multipart()
                    print('recieved',msg)
                    empty = msg.pop(0)
                    assert empty == b''
                    command = msg.pop(0)
                    if command == CREQUEST + CSUCCESS:
                        server = msg.pop(0)
                    #print(server)
                    #print(self.futureobjectdict)
                        msg = pickle.loads(msg[0])
                    #print(msg)
                        self.futureobjectdict[server].set_result(msg)
                    elif command == CHEARTBEAT:
                        answ = [empty,b'S',CHEARTBEAT]
                        self.socket.send_multipart(answ) 
                    elif command == CREQUEST + CFAIL:
                        server = msg.pop(0)
                        self.futureobjectdict[server].set_exception(Exception(msg.pop(0)))

                    #print(msg)
                    toc = time.time()
                    if toc-tic > 1:
                        answ = [empty,b'H',self.name]
                        self.QWeatherStation.send_multipart(answ)    
                        tic = toc
            except KeyboardInterrupt:
                break
                #self.socket.close()
                #self.context.term()



    def __repr__(self):
        #self.get_server_info()
        msg = ""
        lst = [getattr(self,server) for server in dir(self) if getattr(getattr(self,server),'is_remote_server',False)]
        if len(lst) == 0:
            return 'No servers connected'
        else:
            for aserver in lst:
                msg += aserver.name + "\n"
        return msg.strip()


from setuptools import setup, find_packages

setup(
    name='QWeather',
    version='1.0.1',
    packages=['qweather',],
    author='A. Arvad Joergensen',
    author_email='Asbjorn.Arvad@nbi.ku.dk',
    license='MIT',
    url='https://github.com/Arvad/QWeather',
    description='A zeromq based distributed messaging platform',
    long_description=open('README.rst').read(),
    install_requires=[
        "pyzmq >= 17.0.0.b3"
    ],
    rpython_requires='~=3.5'
)


from .constants import *
import zmq
import pickle
import time
#from zmq.asyncio import Context, Poller

class QWeatherStation:

    def __init__(self,IP,loop = None,verbose=False,debug = False):
        if loop is None:
            from zmq import Context,Poller
        else:
            from zmq.asyncio import Context,Poller
        self.StationIP = IP
        self.verbose = verbose
        self.debug = debug
        self.servers = {}
        self.clients = []
        self.cnx = Context()
        self.socket = self.cnx.socket(zmq.ROUTER)
        self.poller = Poller()
        self.poller.register(self.socket,zmq.POLLIN)
        self.socket.bind(self.StationIP)
        self.heartbeatlist = {}
        self.heartbeattimeout = 2
        if self.verbose:
            print('QWeatherStation ready to run on IP: "',self.StationIP,'"')

    async def async_run(self):
        while True:
            try:
                items = await self.poller.poll(1000)
            except KeyboardInterrupt:
                break

            if items:
                msg = await self.socket.recv_multipart()
                self.handle_message(msg)

    def run(self):
        tic = time.time()
        while True:
            try:
                items = self.poller.poll(1000)
            except KeyboardInterrupt:
                break

            if items:
                msg = self.socket.recv_multipart()
                self.handle_message(msg)
            toc = time.time()
            if toc-tic > self.heartbeattimeout:
                tic = time.time()
                for aconnection in self.heartbeatlist.items():
                    self.heartbeatlist[aconnection[0]] -= 1
                for aconnection in self.heartbeatlist.items():
                    if aconnection[1] < 1:
                        if self.debug:
                            print('Removing ',aconnection[0],' due to timeout')
                        try:
                            self.clients.remove(aconnection[0])
                        except ValueError:
                            pass
                        try:
                            self.servers.pop(aconnection[0])
                        except KeyError:
                            pass
                self.heartbeatlist = dict((k,v) for k,v in self.heartbeatlist.items() if v>0)

        

    def handle_message(self,msg):
        sender = msg.pop(0)
        if self.debug:
            print('DEBUG: From "',sender,'": ',msg)
        empty = msg.pop(0)
        assert empty == b''
        SenderType = msg.pop(0)
        if SenderType == b'S': #server
            command = msg.pop(0) # 0xFx for server and 0x0x for client
            self.process_server(sender,command,msg)
        elif (SenderType == b'C'): #client
            command = msg.pop(0) # 0xFx for server and 0x0x for client
            self.process_client(sender,command,msg)

        elif SenderType == b'H': #heartbeat
            sender = msg.pop(0).decode()
            if self.debug:
                print('Recieved heartbeat from ',sender)
            if sender in self.heartbeatlist:
                self.heartbeatlist[sender] = HEARTBEATMAX
        else:
            if self.verbose:
                print('Invalid message')

    def process_client(self,sender,command,msg):
        if command == CREADY:
            version = msg.pop(0)
            if not version == PCLIENT:
                newmsg = [sender,b'',CREADY + CFAIL,'Mismatch in protocol between client and broker'.encode()]
            else:
                newmsg = [sender,b'',CREADY + CSUCCESS] + [pickle.dumps(self.servers)]

                if self.verbose:
                    print('Client ready at "',int.from_bytes(sender,byteorder='big'),'"')
                name = msg.pop(0)
                self.clients.append(name.decode())
                self.heartbeatlist[name.decode()] =HEARTBEATMAX
            self.socket.send_multipart(newmsg)

        elif command == CREQUEST:
            server = msg.pop(0).decode()
            serveraddr = self.servers[server][0]
            msg = [serveraddr,b'',CREQUEST,sender] + msg
            print(len(self.servers[server][2]))
            if len(self.servers[server][2]) ==  0:
                self.socket.send_multipart(msg)
                if self.debug:
                    print('DEBUG: CLient request at"',sender,'":',msg)
            else:
                self.servers[server][2].append(msg)



    def process_server(self,sender,command,msg):
        if command == CREADY:
            version = msg.pop(0)
            if not version == PSERVER:
                print('went here')
                newmsg = [sender,b'',CREADY + CFAIL,'Mismatch in protocol between server and broker'.encode()]
            else:
                servername = msg.pop(0).decode()
                servermethods = pickle.loads(msg.pop(0))
                self.servers[servername] = (sender,servermethods,[])
                newmsg = [sender,b'',CREADY + CSUCCESS]
                self.heartbeatlist[servername] = HEARTBEATMAX
                if self.verbose:
                    print('Server "',servername,'" ready at: "',int.from_bytes(sender,byteorder='big'),'"')
            self.socket.send_multipart(newmsg)

        elif command == CREPLY:
            server = msg.pop(0).decode()
            client = msg.pop(0)
            answ = msg.pop(0)
            msg = [client,b'',CREQUEST + CSUCCESS,server.encode(),answ]
            self.socket.send_multipart(msg)
            if self.debug:
                print('DEBUG: To "',client,'"',msg)
            if len(self.servers[server][2]) > 0:
                self.socket.send_multipart(self.servers[server][2].pop(0))
                if self.debug:
                    print('DEBUG: CLient request at"',sender,'":',msg)

